class UniError(Exception):
    pass


class UniAnswerDelayError(UniError):
    pass


class UniRedundantAnswerError(UniError):
    pass


class UniEmptyAnswerError(UniError):
    pass


class UniConfigError(UniError):
    pass


class UniDefinitionNotFoundError(UniError):
    pass


class UniMessageError(UniError):
    pass


class UniMessagePayloadParsingError(UniMessageError):
    pass


class UniAnswerMessagePayloadParsingError(UniMessageError):
    pass


class UniPayloadSerializationError(UniMessageError):
    pass


class UniSendingToUndefinedWorkerError(UniError):
    pass


class UniMessageRejectError(UniError):
    pass



from typing import Callable, Union, Any, Optional, Dict, TypeVar
from uuid import uuid4, UUID

from mypy_extensions import NamedArg

from unipipeline.answer.uni_answer_message import UniAnswerMessage
from unipipeline.message.uni_message import UniMessage
from unipipeline.worker.uni_msg_params import UniSendingParams, UniGettingAnswerParams, default_getting_answer_params, default_sending_params, TUniSendingMessagePayloadUnion

TInputMessage = TypeVar('TInputMessage', bound=UniMessage)
TAnswMessage = TypeVar('TAnswMessage', bound=UniMessage)


class UniWorkerConsumerManager:
    def __init__(
        self,
        send: Callable[[str, TUniSendingMessagePayloadUnion, NamedArg(UniSendingParams, 'params')], None],
        get_answer_from: Callable[[str, TUniSendingMessagePayloadUnion, NamedArg(UniGettingAnswerParams, 'params')], Optional[UniAnswerMessage[UniMessage]]],
    ) -> None:
        self._send = send
        self._get_answer_from = get_answer_from
        self._id = uuid4()

    @property
    def id(self) -> UUID:
        return self._id

    def stop_consuming(self) -> None:
        raise NotImplementedError(f'{type(self).__name__}.stop_consuming was not implemented')  # TODO

    def exit(self) -> None:
        raise NotImplementedError(f'{type(self).__name__}.exit was not implemented')  # TODO

    def get_answer_from(self, worker: str, data: Union[Dict[str, Any], UniMessage], params: UniGettingAnswerParams = default_getting_answer_params) -> Optional[UniAnswerMessage[TAnswMessage]]:
        return self._get_answer_from(worker, data, params=params)  # type: ignore

    def send_to(self, worker: str, data: Union[Dict[str, Any], TInputMessage], params: UniSendingParams = default_sending_params) -> None:
        self._send(worker, data, params=params)



from unipipeline.definitions.uni_definition import UniDefinition
from unipipeline.definitions.uni_module_definition import UniModuleDefinition


class UniCodecDefinition(UniDefinition):
    name: str
    encoder_type: UniModuleDefinition
    decoder_type: UniModuleDefinition



import signal
from types import FrameType
from typing import Callable, Tuple, Any, Optional, Union


class soft_interruption:  # noqa
    def __init__(self, on_interrupt_once: Callable[[], None], on_force_interruption: Optional[Callable[[], None]] = None, on_error: Optional[Callable[[Exception], None]] = None) -> None:
        self._on_interrupt_once = on_interrupt_once
        self._on_force_interruption = on_force_interruption
        self._sig_int_received = False
        self._sig_term_received = False
        self._first_sig_data: Optional[Tuple[Any, Any]] = None
        self._old_sig_int: Union[Callable[[signal.Signals, FrameType], Any], int, signal.Handlers, None] = None
        self._old_sig_term: Union[Callable[[signal.Signals, FrameType], Any], int, signal.Handlers, None] = None
        self._on_error = on_error

    def _the_end(self, force: bool = False) -> None:
        if force:
            self._on_force_interruption and self._on_force_interruption()
        if self._sig_int_received and callable(self._old_sig_int):
            assert self._first_sig_data is not None
            self._old_sig_int(*self._first_sig_data)
        elif self._sig_term_received and callable(self._old_sig_term):
            assert self._first_sig_data is not None
            self._old_sig_term(*self._first_sig_data)
        if force or self._sig_term_received or self._sig_int_received:
            exit(1)

    def _interruption_handler(self, sig: signal.Signals, frame: FrameType) -> None:
        try:
            if self._sig_term_received or self._sig_int_received:
                self._the_end(True)
                return
            self._sig_term_received = True
            self._first_sig_data = (sig, frame)
            self._on_interrupt_once()
        except Exception as e:  # noqa
            if self._on_error:
                self._on_error(e)
            else:
                print(e)  # noqa

    def _term_handler(self, sig: signal.Signals, frame: FrameType) -> None:
        try:
            if self._sig_term_received or self._sig_int_received:
                self._the_end(True)
                return
            self._sig_term_received = True
            self._first_sig_data = (sig, frame)
            self._on_interrupt_once()
        except Exception as e:  # noqa
            if self._on_error:
                self._on_error(e)
            else:
                print(e)  # noqa

    def __enter__(self) -> None:
        self._old_sig_int = signal.signal(signal.SIGINT, self._interruption_handler)  # type: ignore
        self._old_sig_term = signal.signal(signal.SIGTERM, self._term_handler)  # type: ignore
        return None

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        self._the_end(False)



import json
from datetime import datetime
from enum import Enum
from typing import Any
from uuid import UUID


class ComplexEncoder(json.JSONEncoder):
    def default(self, obj: Any) -> Any:
        if isinstance(obj, datetime):
            return obj.isoformat()
        if isinstance(obj, UUID):
            return str(obj)
        if isinstance(obj, bytes):
            return obj.hex()
        if isinstance(obj, Enum):
            return obj.value
        return json.JSONEncoder.default(self, obj)


def complex_serializer_json_dumps(data: Any) -> str:
    return ComplexEncoder().encode(data)



import importlib
import os.path
from time import sleep
from typing import NamedTuple, Type, TypeVar, Any, Optional, Set, Callable

from unipipeline.utils.uni_echo import UniEcho
from unipipeline.utils.uni_util import UniUtil

T = TypeVar('T')

uni_broker_template = '''
from typing import Set, List

from unipipeline import UniBroker, UniMessageMeta, UniBrokerConsumer


class {{name}}(UniBroker):

    def stop_consuming(self) -> None:
        raise NotImplementedError('method stop_consuming must be specified for class "{{name}}"')

    def connect(self) -> None:
        raise NotImplementedError('method connect must be specified for class "{{name}}"')

    def close(self) -> None:
        raise NotImplementedError('method close must be specified for class "{{name}}"')

    def add_consumer(self, consumer: UniBrokerConsumer) -> None:
        raise NotImplementedError('method add_topic_consumer must be specified for class "{{name}}"')

    def start_consuming(self) -> None:
        raise NotImplementedError('method start_consuming must be specified for class "{{name}}"')

    def publish(self, topic: str, meta_list: List[UniMessageMeta]) -> None:
        raise NotImplementedError('method publish must be specified for class "{{name}}"')

    def get_topic_approximate_messages_count(self, topic: str) -> int:
        raise NotImplementedError('method get_topic_approximate_messages_count must be specified for class "{{name}}"')

    def initialize(self, topics: Set[str]) -> None:
        raise NotImplementedError('method initialize must be specified for class "{{name}}"')

'''

uni_message_template = '''from unipipeline import UniMessage


class {{name}}(UniMessage):
    pass

'''


uni_worker_template = '''from unipipeline import UniWorker, UniWorkerConsumerMessage

from {{data.input_message.type.module}} import {{data.input_message.type.object_name}}


class {{name}}(UniWorker[{{data.input_message.type.object_name}}, None]):
    def handle_message(self, msg: UniWorkerConsumerMessage[{{data.input_message.type.object_name}}]) -> None:
        raise NotImplementedError('method handle_message must be specified for class "{{name}}"')

'''


uni_waiting_template = '''from unipipeline import UniWaiting


class {{name}}(UniWaiting):
    def try_to_connect(self) -> None:
        raise NotImplementedError('method try_to_connect must be specified for class "{{name}}"')

'''

tpl_map = {
    "UniBroker": uni_broker_template,
    "UniMessage": uni_message_template,
    "UniWorker": uni_worker_template,
    "UniWaiting": uni_waiting_template,
}


CWD = str(os.getcwdb().decode('utf-8'))


class UniModuleDefinition(NamedTuple):
    module: str
    object_name: str

    @staticmethod
    def parse(type_def: str) -> 'UniModuleDefinition':
        assert isinstance(type_def, str), f"type_def must be str. {type(type_def)} given"
        spec = type_def.split(":")
        assert len(spec) == 2, f'must have 2 segments. {len(spec)} was given from "{type_def}"'
        return UniModuleDefinition(
            module=spec[0],
            object_name=spec[1],
        )

    def import_function(self) -> Callable[..., Any]:
        mdl = importlib.import_module(self.module)
        tp = getattr(mdl, self.object_name)
        if not callable(tp):
            raise TypeError(f'object "{self.module}:{self.object_name}" is not callable')
        return tp  # type: ignore

    def import_class(self, class_type: Type[T], echo: UniEcho, auto_create: bool = False, create_template_params: Any = None, util: Optional[UniUtil] = None) -> Type[T]:
        try:
            mdl = importlib.import_module(self.module)
        except ModuleNotFoundError:
            if not auto_create:
                raise

            if util is None:
                raise TypeError('util must be set')

            mdl = None
            echo = echo.mk_child(f'module[{self.module}::{self.object_name}]')
            hierarchy = self.module.split('.')
            path = os.path.abspath(f'{os.path.join("../modules/", *hierarchy)}.py')
            path_dir = os.path.dirname(path)
            os.makedirs(path_dir, exist_ok=True)

            path_inits: Set[str] = {os.path.join(path_dir, "../modules/__init__.py"), }

            if path_dir.startswith(CWD):
                current_dir: Optional[str] = None
                for pi_dir in path_dir[len(CWD):].strip('/').split('/'):
                    if current_dir is not None:
                        pi_dir = os.path.join(current_dir, pi_dir)
                    current_dir = pi_dir
                    path_inits.add(os.path.join(CWD, pi_dir, "../modules/__init__.py"))

            for pi in path_inits:
                if not os.path.isfile(pi):
                    with open(pi, "wt+") as fi:
                        fi.write("")
                    echo.log_debug(f'file {pi} was created')

            with open(path, 'wt') as fm:
                fm.writelines(util.template.template(tpl_map[class_type.__name__], data=create_template_params, name=self.object_name))
                echo.log_info(f'file {path} was created')

            success = False
            for i in range(10):  # because fs has cache time
                try:
                    importlib.invalidate_caches()
                    mdl = importlib.import_module(self.module)
                    success = True
                    break
                except ModuleNotFoundError:
                    echo.log_debug(f'still not found. try to waiting for {i}s before invalidating the cache')
                    sleep(i)
            if not success:
                echo.log_error('could not be loaded')
                exit(1)
        assert mdl is not None
        tp = getattr(mdl, self.object_name)
        if not issubclass(tp, class_type):
            ValueError(f'class {self.object_name} is not subclass of {class_type.__name__}')
        return tp  # type: ignore



import math
import uuid
from typing import TypeVar, Generic, Optional, Type, Any, Union, Dict, TYPE_CHECKING, Callable

from unipipeline.answer.uni_answer_message import UniAnswerMessage
from unipipeline.definitions.uni_worker_definition import UniWorkerDefinition
from unipipeline.errors import UniMessagePayloadParsingError, \
    UniAnswerMessagePayloadParsingError, UniSendingToUndefinedWorkerError
from unipipeline.message.uni_message import UniMessage
from unipipeline.message_meta.uni_message_meta import UniMessageMeta, UniMessageMetaErrTopic, UniAnswerParams
from unipipeline.worker.uni_msg_params import UniGettingAnswerParams, UniSendingParams, TUniSendingMessagePayloadUnion, TUniSendingWorkerUnion
from unipipeline.worker.uni_worker import UniWorker
from unipipeline.worker.uni_worker_consumer_manager import UniWorkerConsumerManager
from unipipeline.worker.uni_worker_consumer_message import UniWorkerConsumerMessage

if TYPE_CHECKING:
    from unipipeline.modules.uni_mediator import UniMediator

TInputMsgPayload = TypeVar('TInputMsgPayload', bound=UniMessage)
TAnswerMsgPayload = TypeVar('TAnswerMsgPayload', bound=Optional[UniMessage])


class UniWorkerConsumer(Generic[TInputMsgPayload, TAnswerMsgPayload]):
    def __init__(self, definition: UniWorkerDefinition, mediator: 'UniMediator', worker_type: Type[UniWorker[TInputMsgPayload, TAnswerMsgPayload]]) -> None:
        self._definition = definition
        self._mediator = mediator

        self._worker_manager = UniWorkerConsumerManager(self._send_to, self._get_answer_from)

        self._worker = worker_type(self._worker_manager)
        self._uni_echo = mediator.echo.mk_child(f'worker[{definition.name}]')

        self._input_message_type: Type[TInputMsgPayload] = mediator.get_message_type(self._definition.input_message.name)  # type: ignore
        self._answer_message_type: Optional[Type[TAnswerMsgPayload]] = mediator.get_message_type(self._definition.answer_message.name) if self._definition.answer_message is not None else None  # type: ignore

        self._current_meta: Optional[UniMessageMeta] = None

    def _get_answer_from(
        self,
        worker: TUniSendingWorkerUnion,
        data: TUniSendingMessagePayloadUnion,
        *,
        params: UniGettingAnswerParams,
    ) -> Optional[UniAnswerMessage[UniMessage]]:
        answer_params = UniAnswerParams(
            topic=self._definition.answer_topic,
            id=uuid.uuid4(),  # self._worker_manager.id
            ttl_s=math.ceil(params.answer_tll.total_seconds()),
        )
        return self._mediator.get_answer_from(worker, data, parent_meta=self._current_meta, answer_params=answer_params, params=params)

    def _send_to(
        self,
        worker: TUniSendingWorkerUnion,
        data: TUniSendingMessagePayloadUnion,
        *,
        params: UniSendingParams,
    ) -> None:
        wd = self._mediator.config.get_worker_definition(worker)
        if wd.name not in self._definition.output_workers:
            raise UniSendingToUndefinedWorkerError(f'worker {wd.name} is not defined in workers->{self._definition.name}->output_workers')
        self._mediator.send_to(wd.name, data, parent_meta=self._current_meta, params=params)

    def process_message(self, get_meta: Callable[[], UniMessageMeta]) -> None:
        self._current_meta = None

        self._uni_echo.log_debug('processing start')
        try:
            meta = get_meta()
            msg = UniWorkerConsumerMessage[TInputMsgPayload](self._input_message_type, meta)
        except Exception as e: # noqa
            self._uni_echo.log_error(str(e))
            return

        self._current_meta = meta
        self._uni_echo.log_debug(f'meta unpacked successfully {meta.id}')

        try:
            result: Optional[Union[TAnswerMsgPayload, Dict[str, Any]]] = self._worker.handle_message(msg)

        except UniAnswerMessagePayloadParsingError as e:
            self._uni_echo.log_warning(f'payload is invalid! message {meta.id} was skipped')
            self._mediator.move_to_error_topic(self._definition, meta, UniMessageMetaErrTopic.ANSWER_MESSAGE_PAYLOAD_ERR, e)
            self._current_meta = None
            return

        except UniMessagePayloadParsingError as e:
            self._uni_echo.log_warning(f'answer payload is invalid! message {meta.id} was skipped')
            self._mediator.move_to_error_topic(self._definition, meta, UniMessageMetaErrTopic.MESSAGE_PAYLOAD_ERR, e)
            self._current_meta = None
            return

        self._uni_echo.log_debug(f'processing successfully done {meta.id}')
        if meta.need_answer and self._definition.need_answer:
            try:
                self._mediator.answer_to(self._definition.name, meta, result, unwrapped=self._definition.answer_unwrapped)
            except UniSendingToUndefinedWorkerError:
                pass

        self._current_meta = None
